package at.ac.uibk.gitsearch.service;

import java.io.File;
import java.io.FileOutputStream;
import java.io.IOException;
import java.io.InputStream;
import java.net.URI;
import java.net.URISyntaxException;
import java.util.Arrays;
import java.util.List;
import java.util.Optional;
import java.util.stream.Stream;
import java.util.zip.ZipInputStream;

import org.gitlab4j.api.GitLabApi;
import org.gitlab4j.api.GitLabApiException;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.security.core.userdetails.User;
import org.springframework.stereotype.Service;
import org.springframework.web.servlet.support.ServletUriComponentsBuilder;

import at.ac.uibk.gitsearch.repository.gitlab.GitLabRepository;
import at.ac.uibk.gitsearch.repository.search.MetaDataRepository;
import at.ac.uibk.gitsearch.security.jwt.TokenProvider;
import at.ac.uibk.gitsearch.service.dto.AutoCompleteEntry;
import at.ac.uibk.gitsearch.service.dto.SearchInputDTO;
import at.ac.uibk.gitsearch.service.dto.SearchResultDTO;
import at.ac.uibk.gitsearch.service.dto.SearchResultsDTO;

/**
 * Service for exercise/course search results
 * <p>
 */
@Service
public class SearchService {

	private final Logger log = LoggerFactory.getLogger(SearchService.class);

	@Autowired
	protected MetaDataRepository metaDataRepository;
	@Autowired
	protected TokenProvider tokenProvider;
	@Autowired
	protected GitLabRepository gitLabRepository;
	@Autowired
	protected PluginManagementService pluginManagementService;
	@Autowired
	protected ShoppingBasketService shoppingBasketService;

	/** just for test implementation **/
	public static final int NUM_TESTRESULTS = 23;

	public SearchService() {
		// Just for Spring
	}

	/**
	 * returns all keyword autocompletes for keyWord
	 *
	 * @param keyWordPrefix
	 * @return
	 * @throws IOException
	 */
	public List<AutoCompleteEntry> getKeywordsAutoComplete(String keyWordPrefix) throws IOException {
		return metaDataRepository.getKeywordsAutoComplete(keyWordPrefix);
	}

		/**
	 * returns all keyword autocompletes for keyWord
	 *
	 * @param formatPrefix
	 * @return
	 * @throws IOException
	 */
	public List<AutoCompleteEntry> getFormatsAutoComplete(String formatPrefix) throws IOException {
		return metaDataRepository.getFormatsAutoComplete(formatPrefix);
	}

	/**
	 * returns all creator autocompletes
	 *
	 * @param creatorPrefix
	 * @return
	 * @throws IOException
	 */
	public List<AutoCompleteEntry> getCreatorAutoComplete(String creatorPrefix) throws IOException {
		return metaDataRepository.getCreatorAutoComplete(creatorPrefix);
	}

	/**
	 * returns all contributor autocompletes
	 *
	 * @param contributorPrefix
	 * @return
	 * @throws IOException
	 */
	public List<AutoCompleteEntry> getContributorAutoComplete(String contributorPrefix) throws IOException {
		return metaDataRepository.getContributorAutoComplete(contributorPrefix);
	}

	/**
	 * returns all contributor and author autocompletes
	 *
	 * @param contributorPrefix
	 * @return
	 * @throws IOException
	 */
	public List<AutoCompleteEntry> getContributorCreatorAutoComplete(String contributorPrefix) throws IOException {
		return metaDataRepository.getContributorCreatorAutoComplete(contributorPrefix);
	}

	/**
	 * returns all programmingLanguage autocompletes for keyWord
	 *
	 * @param programmingLanguagePrefix
	 * @return
	 * @throws IOException
	 */
	public List<AutoCompleteEntry> getProgrammingLanguageAutoComplete(String programmingLanguagePrefix)
			throws IOException {
		return metaDataRepository.getProgrammingLanguageAutoComplete(programmingLanguagePrefix);
	}

	/**
	 * returns the result page
	 *
	 * @param searchInput the query definition
	 * @param first       the index of the first record to be returned
	 * @param pageLength  the number of records on each page
	 */
	public SearchResultsDTO searchResultPage(org.codeability.sharing.plugins.api.search.SearchInputDTO searchInput) throws IOException {
		final Optional<User> principal = tokenProvider.getCurrentPrincipal();
		log.debug("Searchrequest for {} ", searchInput);
		
		final SearchResultsDTO pageDetails = metaDataRepository.pageDetails(searchInput,
				principal);

		pageDetails.getSearchResult().stream()
				.forEach(hit -> pluginManagementService.getRegisteredPluginConfigs().stream()
						.forEach(config -> config.getActions().values().stream()
								.filter(action -> action.isApplicable(hit))
								.forEach(action -> hit.getSupportedActions().add(action.getPluginActionInfo()))));
		pageDetails.getSearchResult().stream().forEach(this::fixImageURL);

		return pageDetails;
	}

	/**
	 * returns the result page
	 *
	 * @param searchInput the query definition
	 * @param page       the index of the first record to be returned
	 * @param length      the number of records on page
	 */
	public SearchResultsDTO watchListResultPage(Stream<String> exerciseIds, int page, int length) throws IOException {

		final SearchResultsDTO pageDetails = metaDataRepository.getExercisesById(exerciseIds,tokenProvider.getCurrentPrincipal(), page, length);

		pageDetails.getSearchResult().stream()
				.forEach(hit -> pluginManagementService.getRegisteredPluginConfigs().stream()
						.forEach(config -> config.getActions().values().stream()
								.filter(action -> action.isApplicable(hit))
								.forEach(action -> hit.getSupportedActions().add(action.getPluginActionInfo()))));
		pageDetails.getSearchResult().stream().forEach(this::fixImageURL);

		return pageDetails;
		// return readTestResults(searchInput.getFulltextQuery(), first, length);
	}

	private void fixImageURL(SearchResultDTO metaData) {

		String image = metaData.getMetadata().getImage();

		try {
			if (image != null) {
				URI url = new URI(image);
				if (url.isAbsolute())
					return;
				String httpUrlToRepo = metaData.getProject().getUrl();
				String baseRepoURL = httpUrlToRepo + "/-/raw/master/";
				final URI resolvedImageUrl = new URI(baseRepoURL).resolve(url);
				metaData.getMetadata().setImage(resolvedImageUrl.toASCIIString());
				return;
			}

			final String baseUrl = ServletUriComponentsBuilder.fromCurrentContextPath().build().toUriString();
			final boolean isJava = metaData.getMetadata().getProgrammingLanguage() != null && Arrays
					.stream(metaData.getMetadata().getProgrammingLanguage()).anyMatch(s -> s.startsWith("JAVA"))
					&& metaData.hashCode() % 2 == 0; // nur um ein wenig Abwechslung zu bekommen :-)
			final boolean isLatexFormat = metaData.getMetadata().getFormat() != null
					&& Arrays.stream(metaData.getMetadata().getFormat()).anyMatch(s -> s.startsWith("latex"));
			final boolean isPython = metaData.getMetadata().getProgrammingLanguage() != null && Arrays
					.stream(metaData.getMetadata().getProgrammingLanguage()).anyMatch(s -> s.startsWith("python"));
			if (isJava) {
				metaData.getMetadata().setImage(baseUrl + "/content/img/java.png");
				return;
			}
			if (isLatexFormat) {
				metaData.getMetadata().setImage(baseUrl + "/content/img/latex.png");
				return;
			}
			if (isPython) {
				metaData.getMetadata().setImage(baseUrl + "/content/img/python.png");
				return;
			}
			metaData.getMetadata().setImage(baseUrl + "/content/img/gitLab.png");
			return;

		} catch (URISyntaxException e) {
			log.warn("Cannot parse image url {} for {}", metaData.getMetadata().getImage(),
					metaData.getProject().getUrl());
		}
	}

	
	public File exportExercise(long projectId) throws IOException {
		// TODO projectId and exercise Id are mixed up here!
		try{
		final GitLabApi gitLabApi = gitLabRepository.getGitLabApi(tokenProvider.getGitLabAccessInfo());
		InputStream inputFile = shoppingBasketService.rePackageGitLabProjectZip(new ZipInputStream(gitLabApi.getRepositoryApi().getRepositoryArchive((int)projectId, "HEAD", "zip")), "from project " + projectId);
		File file = new File("exercise" + projectId + ".zip");

		return copyInputStreamToFile(inputFile, file);}
		catch(GitLabApiException exception){
			log.error(exception.getMessage());
			return null;
		}
	}

	private File copyInputStreamToFile(InputStream inputStream, File file) throws IOException {

		// append = false
		try (FileOutputStream outputStream = new FileOutputStream(file, false)) {
			int read;
			byte[] bytes = new byte[8192];
			while ((read = inputStream.read(bytes)) != -1) {
				outputStream.write(bytes, 0, read);
			}
			return file;
		}
	}

}
